use std::collections::HashMap;
use std::sync::Arc;

use enum_map::EnumMap;
use serde::{Deserialize, Serialize};

use crate::model::icon_descriptor::IconDescriptor;
use crate::model::position::KnobPosition;
use crate::model::rgb::RGB8WithOptionalA;
use crate::modes;

#[derive(Debug, Deserialize)]
pub struct File {
    pub id: Option<String>,
    pub knobs: HashMap<KnobPosition, Knob>,
}

#[derive(Debug)]
pub struct Page {
    pub id: String,
    pub knobs: EnumMap<KnobPosition, Arc<Knob>>,
}

#[derive(Debug, Default, Deserialize)]
pub struct Knob {
    #[serde(default, flatten)]
    pub base_style: KnobStyle,
    #[serde(default)]
    pub mode: KnobModes,
}

#[derive(Debug, Default, Clone, PartialEq, Serialize, Deserialize)]
pub struct KnobIndicators {
    pub bar: Option<KnobIndicatorBarConfig>,
    pub circle: Option<KnobIndicatorCircleConfig>,
}

impl KnobIndicators {
    pub fn merge_over(&self, base: &Self) -> Self {
        Self {
            bar: self
                .bar
                .as_ref()
                .zip(base.bar.as_ref())
                .map(|(a, b)| a.merge_over(b))
                .or_else(|| self.bar.clone())
                .or_else(|| base.bar.clone()),
            circle: self
                .circle
                .as_ref()
                .zip(base.circle.as_ref())
                .map(|(a, b)| a.merge_over(b))
                .or_else(|| self.circle.clone())
                .or_else(|| base.circle.clone()),
        }
    }
}

#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
pub struct KnobIndicatorBarConfig {
    pub color: Option<RGB8WithOptionalA>,
}

impl KnobIndicatorBarConfig {
    pub fn merge_over(&self, base: &Self) -> Self {
        Self {
            color: self.color.as_ref().or(base.color.as_ref()).cloned(),
        }
    }
}

#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
pub struct KnobIndicatorCircleConfig {
    pub color: Option<RGB8WithOptionalA>,
    pub width: Option<u8>,
    pub radius: Option<u8>,
}

impl KnobIndicatorCircleConfig {
    pub fn merge_over(&self, base: &Self) -> Self {
        Self {
            color: self.color.as_ref().or(base.color.as_ref()).cloned(),
            width: self.width.as_ref().or(base.width.as_ref()).cloned(),
            radius: self.radius.as_ref().or(base.radius.as_ref()).cloned(),
        }
    }
}

#[derive(Debug, Default, Clone, PartialEq, Serialize, Deserialize)]
pub struct KnobStyle {
    pub label: Option<String>,
    pub icon: Option<IconDescriptor>,
    pub indicators: Option<KnobIndicators>,
}

impl KnobStyle {
    pub fn merge_over(&self, base: &Self) -> Self {
        Self {
            label: self.label.as_ref().or(base.label.as_ref()).cloned(),
            icon: self.icon.as_ref().or(base.icon.as_ref()).cloned(),
            indicators: self
                .indicators
                .as_ref()
                .zip(base.indicators.as_ref())
                .map(|(a, b)| a.merge_over(b))
                .or_else(|| self.indicators.clone())
                .or_else(|| base.indicators.clone()),
        }
    }
}

#[derive(Debug, Default, Deserialize)]
pub struct KnobModes {
    pub audio_volume: Option<Arc<modes::knob::audio_volume::Config>>,
}

pub type StyleByStateMap<State> = HashMap<State, KnobStyle>;
