use std::sync::Arc;

use serde::{Deserialize, Serialize};
use tokio::sync::broadcast;

use crate::model;
use crate::model::position::KeyPath;
use crate::runner::command::IoWorkerCommand;

pub mod command;
pub mod home_assistant;
pub mod playerctl;
pub mod timer;
pub mod vibrate;

#[derive(Debug, PartialEq, Clone, Copy, Serialize, Deserialize)]
pub enum KeyTouchEventKind {
    Start,
    Move,
    End,
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
pub enum KeyEvent {
    Press,
    Touch { touch_id: u8, x: u16, y: u16, kind: KeyTouchEventKind },
    VisibilityChange { is_visible: bool },
}

pub fn start_handlers(
    keys: impl Iterator<Item = (KeyPath, Arc<model::key_page::Key>)>,
    events: broadcast::Sender<(KeyPath, KeyEvent)>,
    commands: flume::Sender<IoWorkerCommand>,
) {
    for (path, config) in keys {
        let mut events = events.subscribe();
        let own_events = broadcast::Sender::new(5);

        if let Some(c) = &config.mode.command {
            tokio::spawn(command::handle(Arc::clone(c), own_events.subscribe()));
        }

        if let Some(c) = &config.mode.playerctl__button {
            tokio::spawn(playerctl::handle_button(path.clone(), Arc::clone(c), own_events.subscribe(), commands.clone()));
        }

        if let Some(c) = &config.mode.playerctl__shuffle {
            tokio::spawn(playerctl::handle_shuffle(path.clone(), Arc::clone(c), own_events.subscribe(), commands.clone()));
        }

        if let Some(c) = &config.mode.playerctl__loop {
            tokio::spawn(playerctl::handle_loop(path.clone(), Arc::clone(c), own_events.subscribe(), commands.clone()));
        }

        if let Some(c) = &config.mode.vibrate {
            tokio::spawn(vibrate::handle(Arc::clone(c), own_events.subscribe(), commands.clone()));
        }

        tokio::spawn(async move {
            while let Ok((p, e)) = events.recv().await {
                #[allow(clippy::collapsible_if)]
                if p == path {
                    if own_events.send(e).is_err() {
                        break;
                    }
                }
            }
        });
    }
}
